(* ::Package:: *)

BeginPackage["MCApprox`"]


(* ::Subsection:: *)
(*Usage descriptions*)


EMAlgorithm::usage="";


OptimalCouplingStruct::usage= "";


Bisimilarity::usage= "";


MinimizeMC::usage= "";


MaxIter::usage= "";


Timeout::usage= "";


MC::usage="CTMC[<states>,<transition function>,<labeling function>] is the data structure for a Markov chain. 
The first element of <states> is intened to be the initial state of the chain.";


MCQ::usage="MCQ[<MC>] checks whether <MC> is a well-defined Markov Chain with rational transition function.";


MCtf::usage="MCtf[<array rules>, <states>] returns a transition function out of an array rules describing the transition matrix.";


JoinMC::usage="JoinMC[<MC>,<MC>] returns the disjoint union of two Markov chains.";


PlotMC::usage="PlotMC[<MC>] displays the underlying graph of the Markov chain <MC>.";


BDistMC::usage="BDistMC[<MC>, <discount factor>, <query list> [, Options]] computes the bisimilarity distance between all the pairs of states in <query list>.";


OptimalCoupling::usage="Option for BDistMC (default value False). If set to True, during the computation some information about the current coupling structure is displayed.";


Verbose::usage="Option for BDistMC (default value False). If set to True, during the computation some information about the current coupling structure is displayed.";


Output::usage="Option for BDistMC (default value RulesForm); if set to RulesForm the output distance will di displaied as a set of rules that associate with each pair in the input query its respective distance value; if set to VerboseForm, the output will display some information about the computation.";


ConsistencyCheck::usage="Option for BDistMC (default value True); if set to True, a sanity check of the input is performed before the actual algorithm starts.";


Estimates::usage="Option for BDistMC (default value {}). An over-estimates for the pair of states {s1,s2} can be given in the form '{s1,s2} -> e'. Then, the on-the-fly strategy will consider 'e' as the exact distance value for the pair '{s1,s2}'";


BisimQuotientMC::usage="BisimQuotientMC[<MC>] returns a MC that represents the bisimilarity quotient of the input MC.";


BisimClassesMC::usage="BisimClassesMC[<MC>] returns the partitioning of the states of <MC> w.r.t. bisimilarity.";


RandomCTMC::usage="RandomMC[<states>, <outer-degree>] randomly generates an MC with given out-degree and absorbing-degree.";


Begin["`Private`"]


(* ::Subsection::Closed:: *)
(*Markov Chains (MCs)*)


Support[\[Mu]_]:=\[Mu][[All,1]]; 
ProbabilityQ[\[Mu]_]:=Total[\[Mu][[All,2]]]==1;
ProbabilityTransitionFunctionQ[S_,\[Tau]_]:=And@@(ProbabilityQ@\[Tau]@#&/@S);
ClosedSupport[S_,\[Tau]_]:= And@@(Complement[Support@\[Tau]@#,S]=={}&/@S);


initState[MC[M_,_,_]]:=First[M];


MCQ[MC[S_,\[Tau]_,ell_]]:= ProbabilityTransitionFunctionQ[\[Tau]] && ClosedSupport[S,\[Tau]];
MCQ[_]:= False;


(* ::Text:: *)
(*Disjoint union of two MCs (particularly useful for computing the distance of two MCs)*)


JoinMC[MC[S1_,\[Tau]1_,ell1_],MC[S2_,\[Tau]2_,ell2_]]:=
Module[{S,\[Tau],ell,addindex},
addindex[i_][x_->p_]:={i,x}->p;

S = Join[left/@S1,right/@S2];
\[Tau][{1,x_}]:= Thread[Rule[left/@(\[Tau]1[x][[All,1]]),\[Tau]1[x][[All,2]]]];
\[Tau][{2,x_}]:= Thread[Rule[right/@(\[Tau]2[x][[All,1]]),\[Tau]2[x][[All,2]]]]; 
(*\[Tau][{1_,x_}]:= addindex[1]/@\[Tau]1[x];
\[Tau][{2_,x_}]:= addindex[2]/@\[Tau]2[x];*)
ell[{1,x_}]:= ell1[x]; ell[{2,x_}]:= ell2[x];
MC[S,\[Tau],ell]
];


left[m_]:={1,m}; leftQ[m_]:= MatchQ[m,{1,_}];
right[n_]:={2,n}; rightQ[m_]:= MatchQ[m,{2,_}];


(* ::Text:: *)
(*PlotMC returns a graphical representation of a given Markov chain *)


PlotMC[MC[S_,tau_,ell_]]:=
Module[{f,g,edgelist,vertexlst,panelLabel,vrender},

f[x_][y_->p_]:= Labeled[x->y,N[p]];
vrender[coord_,name_,{w_,h_}]:= {Lighter@Hue[1/ell[name]],Disk[coord,0.1],Black, Scale[Text[Grid[{{InputForm@ell@name}}],coord],0.1]};

(*panelLabel[s_]:= Tooltip[s,Framed@ell[s]];
vertexlst = (# -> Placed[#,Center,panelLabel])&/@S;*)
edgelist= Flatten[(f[#]/@tau[#])&/@S];

Graph[edgelist,VertexShapeFunction->vrender](*,*)
];


(* ::Text:: *)
(*Chop trasition probability function*)


ChopProbabilities[mc:MC[M_,\[Tau]_,ell_]]:=
Module[{ChopTransition},
ChopTransition[u_->p_]:=u->Chop[p];

Do[\[Tau][m]=Select[(ChopTransition/@ \[Tau][m]),#[[2]]>0&],{m,M}];
mc
];


(* ::Subsection::Closed:: *)
(*LP Utilities*)


GetConstrMatrix[inequalities_, vars_]:=
Module[{ConstrRow,rows,consts},
ConstrRow[inequality_]:=
Module[{b,A},
{b,A}=CoefficientArrays[Equal@@inequality,vars];
{A,{-b,Head[inequality]/.{LessEqual->-1,Equal->0,GreaterEqual->1}}}
];
ArrayFlatten/@Thread[ConstrRow/@inequalities]
];


SolveLinearEquations[equalities_,vars_]:=
Module[{A,b},
{A,b}= GetConstrMatrix[equalities,vars];
LinearSolve[A,b[[All,1]]]
];


MinPositiveSolution[equalities_,vars_]:=
Module[{A,b},
{A,b}= GetConstrMatrix[equalities,vars];
LinearProgramming[ConstantArray[1,Length@vars],A,b]
];


(* ::Subsection::Closed:: *)
(*Utilities*)


SafeDivision[x_,_]:= 0 /; x==0;
SafeDivision[x_,y_]:= x/y;


DebugPrint[s___]:=Print[s]/;OptionValue[BDistMC,Verbose];


PrPrint[Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{basic},
basic[value_,index_]:=Style[EpsilonRemoval@value,Red,Bold]/;MemberQ[basicCells,index];
basic[value_,_]:=value;
TableForm[MapIndexed[basic,V,{2}],TableHeadings->{ToString[#[[2]]]&/@mapS,ToString[#[[2]]]&/@mapT}]
];
PrPrint[x_]:=x;


PrintActiveProblems[dist_,coupling_,queryLst_]:=
(Print["Reachable states from ",queryLst, " in the current coupling"];
Print["ActivePairs",ActivePairs[dist,coupling,queryLst]];
Scan[Print[# -> PrPrint[coupling[#]]]&,ActivePairs[dist,coupling,queryLst]])/;OptionValue[BDistMC,Verbose];


(* ::Subsection::Closed:: *)
(*Transportation Problem*)


scheduleCost[cost_,schedule:Schedule[_,_,basicCells_,V_]]:=
Total[Extract[cost,#]*EpsilonRemoval[Extract[V,#]]&/@ basicCells];


cell2pair[{x_,y_},Schedule[mapS_,mapT_,_,_]]:={x/.mapS,y/.mapT};


getScheduleValue[Schedule[mapS_,mapT_,basicCells_,V_],{s_,t_}]:=
Module[{flip,cell},flip[Rule[x_,y_]]:=Rule[y,x];
(* select the cell corresponding to the pair *)
cell= {s/.(flip/@mapS),t/.(flip/@mapT)};
If[MemberQ[basicCells,cell],(* check if the cell is basic*)
	EpsilonRemoval[Extract[V,cell]],
	0
]
];
getScheduleValue[Undefined,_]:=Abort[];


(* ::Text:: *)
(*Given a coupling C a two pairs of states (s,t) and (u,v) it returns C(s,t)(u,v)*)


getCouplingValue[coupling_,st:{_,_},uv:{_,_}]:=getScheduleValue[coupling[st],uv];


(* ::Text:: *)
(*Given two \[Epsilon]-numbers x and y, EpsilonLeq[x,y] determines if x <= y provided that \[Epsilon] -> +0.*)


EpsilonLeq[x_,y_]:=
If[EpsilonRemoval[x-y]==0,(* check if x-y depends only on \[Epsilon]*)
EpsilonFix[x-y,1]<= 0,(* set \[Epsilon] > 0 and check again *)
EpsilonRemoval[x-y]<= 0 (* set e[]= 0 and check again *)
];


EpsilonMin[x_,y_]:= If[EpsilonLeq[x,y],x,y];


EpsilonPerturbation[\[Pi]s_,\[Pi]t_]:=
Module[{AddEpsilon},
AddEpsilon[k_][{a_,id_}]:={a + k e[],id};
{AddEpsilon[1]/@\[Pi]s, Append[Most[\[Pi]t],AddEpsilon[Length@\[Pi]s][Last@\[Pi]t]]}
];


EpsilonFix[x_,\[Epsilon]value_]:=x/.e[]->\[Epsilon]value;


EpsilonRemoval[x_]:=EpsilonFix[x,0];


indexed[lst_]:=Thread[{lst,Table[i,{i,1,Length@lst}]}];


GuessVertexAux[_][\[Pi]s_,\[Pi]t_]:={}/;\[Pi]s=={}||\[Pi]t=={};
GuessVertexAux[strategy_][\[Pi]s_,\[Pi]t_]:=
Module[{G},
G[{i_,j_}]:=
Module[{rec},
rec=If[EpsilonLeq[\[Pi]s[[i,1]],\[Pi]t[[j,1]]],(*\[Pi]s\[LeftDoubleBracket]i,1\[RightDoubleBracket] \[LessEqual] \[Pi]t\[LeftDoubleBracket]j,1\[RightDoubleBracket]*)
GuessVertexAux[strategy][Drop[\[Pi]s,{i}],ReplacePart[\[Pi]t,{j,1}->\[Pi]t[[j,1]]- \[Pi]s[[i,1]]]],
GuessVertexAux[strategy][ReplacePart[\[Pi]s,{i,1}->\[Pi]s[[i,1]]- \[Pi]t[[j,1]]],Drop[\[Pi]t,{j}]]];
Prepend[rec,{\[Pi]s[[i,2]],\[Pi]t[[j,2]]}->EpsilonMin[\[Pi]s[[i,1]],\[Pi]t[[j,1]]]]
];
G[strategy[\[Pi]s,\[Pi]t]]
];


NorthWestStrategy[_,_]:={1,1};


ToMatrix[\[Pi]s_,\[Pi]t_,vertex_]:=
Module[{mapS,revmapS,mapT,revmapT,toMap,mapIndex,mappedvertex},
toMap[indexlst_]:=
Module[{index=Table[i,{i,1,Length@indexlst}] },
{Thread[Rule[index,indexlst]],Thread[Rule[indexlst,index]]}];
mapIndex[{s_,t_}->val_]:={s/.revmapS,t/.revmapT}->val;

{{mapS,revmapS},{mapT,revmapT}} = toMap[#[[All,2]]]&/@{\[Pi]s,\[Pi]t};
mappedvertex= mapIndex/@vertex;
Schedule[mapS,mapT,mappedvertex[[All,1]],SparseArray[mappedvertex,Length/@{\[Pi]s,\[Pi]t}]]
];


GuessVertex[\[Pi]s_,\[Pi]t_]:=
Module[{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t},
{\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t}=EpsilonPerturbation[\[Pi]s, \[Pi]t];
ToMatrix[\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t,GuessVertexAux[NorthWestStrategy][\[Epsilon]\[Pi]s,\[Epsilon]\[Pi]t]]
];


(* ::Text:: *)
(*AddtoBase[d, vertex] == {reducedcost, non-basic cell} *)
(*returns the non-basic cell which has the minimal reduced cost, if the vertex has no non-basic cells, then it returns {\[Infinity], {0,0}}.*)


AddtoBase[costMatrix_,Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{sizeS=Length@mapS, sizeT=Length@mapT, 
t,getCost,ReducedCost,uv,makeRow,base},

makeRow[{1,j_}]:={SparseArray[{{1,sizeS+j}->1},{1,sizeS+sizeT}]};
makeRow[{i_,j_}]:={SparseArray[{{1,i}->1,{1,sizeS+j}->1},{1,sizeS+sizeT}]};

getCost[{i_,j_}]:=costMatrix[[i,j]];

ReducedCost[_,{i_,j_}]:=
Module[{reducedcost=Chop[getCost[{i,j}]-uv[[i]]-uv[[sizeS+j]],10^-5]},
If[base[[1]]>reducedcost,
base={reducedcost,{i,j}} (* update the current least reduced cost *)
];reducedcost];

(* u-v modifiers (Subscript[u, 1],...,Subscript[u, sizeS],Subscript[v, 1],...,Subscript[v, sizeT]) *)
uv=LinearSolve[Normal@ArrayFlatten[makeRow/@basicCells],getCost/@basicCells];
(*Print["u-v Problem:",MatrixForm@ArrayFlatten[makeRow/@basicCells],
MatrixForm@uv,MatrixForm[getCost/@basicCells]];*)
base={\[Infinity],{0,0}};(* init cell *)
(* scan the cells computing their reduced cost *)
MapIndexed[ReducedCost,V,{2}];
base (* return the cell with the least reduced cost *)
];


(* ::Text:: *)
(*SteppingStone[non-basic cell, vertex] returns the vertex obtained by inserting the given non-basic cell into the base *)


SteppingStone[{s_,t_},vertex:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{M,d,oldBasicCell,newBasicCells},
{oldBasicCell,M}=SteppingStonePath[{s,t},vertex];
(*DebugPrint[MatrixForm@M];*)
(*DebugPrint["old basic cell: ",oldBasicCell, ", new basic cell: ",{s,t}];*)
d= Extract[V,oldBasicCell];
(*DebugPrint["ammount moved: ",d];*)
newBasicCells = basicCells/.oldBasicCell->{s,t};
(*DebugPrint["Old basic cells:",basicCells,", new basic cells",newBasicCells];*)
Schedule[mapS,mapT,newBasicCells,SparseArray[V+d M]]
];


SteppingStonePath[{s_,t_},vertex:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{neighbors,visit,parent,cell,path,getMatrix,visited,f},

visited[_]:=False;
neighbors[{u_,True}]:={#[[2]],False}&/@(Select[basicCells,#[[1]]==u \[And] Not[visited[{#[[2]],False}]]&]);
neighbors[{v_,False}]:={#[[1]],True}&/@(Select[basicCells,#[[2]]==v \[And] Not[visited[{#[[1]],True}]]&]);

f[{{a_,True},{b_,False}}]:={a,b};
f[{{a_,False},{b_,True}}]:={b,a};

visit[node_]:=
(visited[node]=True;
Scan[(parent[#]=node; visit@#)&,neighbors[node]]
);

getMatrix[{},sign_]:={};
getMatrix[lst_,sign_]:=
(If[sign<0\[And] EpsilonLeq[Extract[V,First@lst] ,Extract[V,cell]],cell=First@lst];(* update cell *)
Append[getMatrix[Rest@lst,- sign], First@lst -> sign]); (* compute arrayrules *)

(*DebugPrint[PrPrint@vertex];*)

(* search for an alternating path from s to t *)
parent[{s,True}]={};(* {s,True} is the root of the tree *)
visit[{s,True}]; (* compute the spanning tree rooted in {s,True} *)

(* get the alternating path from {t,False} to {s,True} *)
path=Append[f/@Partition[(Most@NestWhileList[parent,{t,False},#!={}&]),2,1],{s,t}];
cell=First@path; path=getMatrix[path,-1];

{cell,SparseArray[path,Dimensions@V]}
];


BestVertex[costMatrix_,{redCost_,nbCell_},vertex_]:=
Module[{currVertex=vertex, currRedCost=redCost,currNbCell= nbCell},
While[currRedCost<0,
currVertex= SteppingStone[currNbCell,currVertex];(* move to a better vertex *)
(*DebugPrint[PrPrint@currVertex];*)
(* update the current least reduced cost *)
{currRedCost,currNbCell}= AddtoBase[costMatrix,currVertex];
];
currVertex
];


(* ::Subsection::Closed:: *)
(*On-the-fly Algorithm for computing Bisimilarity Pseudometric (exact)*)


FixedDistanceQ[dist_,pair:{_,_}]:= MatchQ[dist@@(Sort@pair),_Fixed] ;(* says if the distance is fixed or not *)


VisitedPairQ[dist_,coupling_,pair:{_,_}]:= (* decide whether the distance pair has already been considered *)
FixedDistanceQ[dist,pair] \[Or] Not@MatchQ[coupling[pair],Undefined];


AddScheduleAt[mc:MC[_,\[Tau]_,_]][coupling_,pair:{s_,t_}]:=
Module[{Marginal,newschedule},
Marginal[state_]:= Cases[\[Tau][state],HoldPattern[index_->v_/;v>0]->{v,index}];
(* guess a fesible TP schedule for each action and set it at 'pair' *)
coupling[pair]= GuessVertex[Marginal[s],Marginal[t]];
coupling
];


CompleteCoupling[mc:MC[_,_,_],coupling_,queryLst_]:= 
Module[{f},
f[coupl_,pair_]:= AddScheduleAt[mc][coupl,pair] /;MatchQ[coupl[pair],Undefined];
f[coupl_,_]:=coupl;

Fold[f,coupling,queryLst]
];


NextDemandedPairs[schedule:Schedule[mapS_,mapT_,basicCells_,V_]]:= 
DeleteDuplicates[Sort[cell2pair[#,schedule]]& /@ Select[basicCells,EpsilonRemoval[Extract[V,#]]>0&]];


AddPair[MC[_,_,_],dist_][coupling_,pair:{_,_}]:= coupling/; VisitedPairQ[dist,coupling,pair];
AddPair[mc:MC[_,\[Tau]_,_],dist_][coupling_,pair:{s_,t_}]:=
Module[{Marginal,newschedule},
Marginal[state_]:= Cases[\[Tau][state],HoldPattern[index_->v_/;v>0]->{v,index}];
(* guess a fesible TP schedule for each action *)
newschedule := GuessVertex[Marginal[s],Marginal[t]];
(* recursively update the coupling *)
UpdateCoupling[mc,dist][coupling,pair,newschedule]
];


AddPairs[mc:MC[_,_,_],dist_][coupling_,queryLst_]:=Fold[AddPair[mc,dist],coupling,queryLst];


UpdateCoupling[mc:MC[_,_,_],dist_][coupling_,pair:{_,_},schedule_Schedule]:=
(coupling[pair]=schedule; 
DebugPrint[NextDemandedPairs[schedule]];
AddPairs[mc,dist][coupling,NextDemandedPairs[schedule]]);


ReachableProblems[dist_,coupling_,queryLst_]:=
Module[{visitedQ}, visitedQ[_]:= False; Fold[DFS[dist,coupling],visitedQ,queryLst]];
DFS[dist_,coupling_][visitedQ_,pair:{_,_}]:=visitedQ /; visitedQ@pair \[Or] FixedDistanceQ[dist,pair];
DFS[dist_,coupling_][visitedQ_,pair:{_,_}]:=
(visitedQ[pair]=True; Fold[DFS[dist,coupling],visitedQ,NextDemandedPairs[coupling[pair]]]);


SelectVisited[visitedQ_]:= #[[1,1,1]]&/@Select[DownValues[Evaluate@visitedQ],#[[2]]&];


ActivePairs[dist_,coupling_,queryLst_]:=SelectVisited@ReachableProblems[dist,coupling,queryLst];
(*#[[1,1,1]]&/@Select[DownValues[Evaluate@ReachableProblems[dist,coupling,queryLst]],#[[2]]&];*)


(* ::Text:: *)
(*UpdateDistanceValues computes the discrepancy associated with the given coupling and updates the distance accordigly*)


UpdateDistanceValues[\[Lambda]_,mc:MC[_,_,_],dist_,coupling_,queryLst_]:=
Module[{discrepancyValues,updateVal},

updateVal[d_,pair:{s_,t_} -> val_]:=(d[s,t]=Fixed[0];(*Unset[coupling[pair]];*)d)/;val==0;
updateVal[d_,pair:{s_,t_} -> val_]:=(d[s,t]=val;d);

(* update the distances with new pair values *)
discrepancyValues = Discrepancy[\[Lambda],mc,dist,coupling,queryLst];
DebugPrint["Current update:",discrepancyValues];
dist=Fold[updateVal,dist,discrepancyValues];
Return[dist]
];


(* ::Text:: *)
(*Compute the dicrepancy associated to a given coupling*)


(* ::Text:: *)
(*Remark: Discrepancy is missing the part that fixes the current bisimilation (reverse reachability of the coupling structure)*)


Discrepancy[\[Lambda]_,MC[_,_,ell_],dist_,coupling_,queryLst_]:=
Module[{\[CapitalGamma],\[CapitalGamma]coupl,activeLst,xVars,xVals,x},

(* d[s,t] \[Equal] \!\(
\*SubsuperscriptBox[\(\[CapitalGamma]\), \(\[Lambda]\), \(C\)]\((d)\)\((s, t)\)\) *)
\[CapitalGamma][pair:{s_,t_}]:= {x@@pair == 1}/;ell[s] != ell[t]; (* d[s,t] \[Equal] 1 if \[ScriptL](s) \[NotEqual] \[ScriptL](t)*)
\[CapitalGamma][pair:{_,_}]:= {x@@pair == \[Lambda] \[CapitalGamma]coupl[coupling[pair]]};(* d[s,t] \[Equal] \[Lambda] Underscript[\[Sum], u,v]C(s,t)(u,v) d[u,v]  otherwise *)

\[CapitalGamma]coupl[schedule:Schedule[_,_,basicCells_,V_]]:=
Module[{var},
var[cell:{_,_}]:= Module[{pair=cell2pair[cell,schedule]},
	If[FixedDistanceQ[dist,pair],getDistValue[dist,pair],x@@(Sort@pair)]];
Total[var[#]*EpsilonRemoval[Extract[V,#]]&/@basicCells]
];

(* get the list of active pairs *)
activeLst=ActivePairs[dist,coupling,queryLst];
(* construct the linear system of equations that charcterizes 
 the discrepancy assiciated with the given coupling *)
xVars = x@@#&/@activeLst;(* value variables *)
DebugPrint[Flatten[\[CapitalGamma]/@activeLst]];
xVals = MinPositiveSolution[Cases[Flatten[\[CapitalGamma]/@activeLst],Except[True|False]],xVars]; (* discrepancy values *)
Thread[Rule[activeLst,xVals]]
];


EmptyCoupling[]:= Module[{C},C[_]:=Undefined; Return[C]];


getDistValue[dist_,pair:{_,_}]:=
Module[{getValue},
getValue[Fixed[v_]]:=v;
getValue[v_]:=v;

getValue[dist@@(Sort@pair)]
];


InitDistances[mc:MC[_,_,ell_],estimates_]:=
Module[{dist,diffIndex,fixEstimate,TOP=1},

dist[s_,s_]:= Fixed[0]; (* {s,s} pairs are fixed to 0 *)
dist[s_,t_]:= Fixed[TOP] /; ell[s] != ell[t]; (* pairs are fixed to 1 *)
dist[_,_]:= TOP;

fixEstimate[pair:{s_,t_}->v_]:= (dist[s,t]=Fixed[v]) /; v >= 0 && v <= 1 && Not@FixedDistanceQ[dist,pair];
Scan[fixEstimate,estimates]; (* set all the given estimates as fixed *)
Return[dist]
];


RelativeCostMatrix[dist_,schedule:Schedule[mapS_,mapT_,basicCells_,V_]]:=
Module[{getCellCost,cost,max},
getCellCost[_,cell_] := getDistValue[dist,cell2pair[cell,schedule]];
MapIndexed[getCellCost,V,{2}]
];


(* ::Text:: *)
(*MatchingToChange[coupling,dist,{s,t}] == {{u,v}, {redCost, nbcell}} if the schedule at (u,v) can be further minimized wrt TP with costs given by dist, otherwise MatchingToChange[coupling,dist,{s,t}] == {}.*)


MatchingToChange[coupling_,dist_,queryLst_]:=
Module[{Q,p,a,reducedcost,schedule,cell,siblings,visited},

visited[pair:{_,_}]:= FixedDistanceQ[dist,pair];

(* a new problem should not be visited if it has been visited already or 
its corresponding value is fixed *)
Q=Select[queryLst,Not@FixedDistanceQ[dist,#]&];
While[Length@Q >0,
(* take a problem from the queue and set it as already visited *)
p = First@Q; visited[p] = True;
(*DebugPrint["Visited problem:",{i,j}];*)
(* compute the least reduced cost for the current problem *)
schedule = coupling[p];
{reducedcost,cell} = AddtoBase[RelativeCostMatrix[dist,schedule],schedule];
If[reducedcost<0,
(* the first encoutered vertex to be changed *)
DebugPrint["Non optimal schedule for: ",p,
" reduced cost: ",reducedcost, " for introducing ",cell," as basic"];
Return[{p,{reducedcost,cell}}],
(* otherwise insert its siblings to the queue *)
siblings = Select[NextDemandedPairs[coupling[p]],Composition[Not,visited]];
Q = DeleteDuplicates@Join[Rest@Q,siblings]
]];
(* none of the reached problems can be further decreased *)
{}
];


Options[BDistMC]={Verbose->False,ConsistencyCheck->True,Output->RulesForm,Estimates-> {}};
BDistMC::query="The following queries are ill-formed: `1`.";
BDistMC::MarkovChain = "The 1st argument is not a well-formed rational-valued MDP.";
BDistMC::DiscountFactor ="The second argument, namely `1`, must be a rational in the interval (0,1]";
BDistMC::MultipleEstimates="There is a conflit between estimates (notice that if you have both 
{s,t}->v1 and {t,s}->v2, v1 must be equal to v2)";
BDistMC::WrongEstimates="Estimates must be a rationals or integers in [0,\[Infinity])";

(*BDistMC[mc:MC[_,_,_],\[Lambda]_, All,
options: OptionsPattern[{RulesForm->True,Verbose -> False,ConsistencyCheck->True,Estimates-> {}}]]:=
BDistMC[ctmc,\[Lambda],AllPairs[mc],options];*)
BDistMC[mc:MC[_,_,_],\[Lambda]_, queryLst_, 
options:OptionsPattern[{Output->RulesForm,Verbose -> False,ConsistencyCheck->False,Estimates-> {}}]]:=
(If[OptionValue[ConsistencyCheck],CheckInput[mc,\[Lambda], queryLst]];
(* Set Options for ComputeDistance *)
SetOptions[BDistMC,Output->OptionValue[Output]];
SetOptions[BDistMC,Verbose->OptionValue[Verbose]];
SetOptions[BDistMC,ConsistencyCheck->OptionValue[ConsistencyCheck]];
SetOptions[BDistMC,Estimates->OptionValue[Estimates]];
(* Start computing the Subscript[d, \[Lambda]](s,t) *)
ComputeDistancesAux[mc,\[Lambda],queryLst]
);


CheckInput[mdp:MDP[_,_,_],\[Lambda]_, queryLst_]:=
Module[{BadQuery,BadEstimate,sizeS=mdpDimensions[mdp][[1]],
badqueries,estimates},

BadQuery[size_][pair:{_Integer,_Integer}]:=Not[And@@(IntervalMemberQ[Interval[{1,size}],#]&/@pair)];
BadQuery[_][_]:=True;
BadEstimate[size_][pair_->value_]:=BadQuery[size][pair] \[Or] value<0\[Or]Not@MatchQ[value,_Rational|_Integer];

estimates = OptionValue[BDistMC,Estimates];
Which[
Not@MarkovChainQ[mdp],
	Message[BDistMC::MarkovChain];Abort[],
\[Lambda]<=0\[Or]\[Lambda]>1\[Or]Not@MatchQ[\[Lambda],_Rational],
	 Message[BDistMC::DiscountFactor,\[Lambda]];Abort[],
Select[estimates,BadEstimate[mdpDimensions[mdp][[1]]]]>0,
	Message[BDistMC::WrongEstimates];Abort[],
Or@@(Length[#]!=1&/@(DeleteDuplicates[#[[All,2]]]&/@GatherBy[estimates,Sort@First@#&])),
	Message[BDistMC::MultipleEstimates];Abort[],
(badqueries =Select[queryLst,BadQuery[sizeS]])>0 ,
	Message[BDistMC::pair,badqueries];Abort[]
];
];


ComputeDistancesAux[mc:MC[S_,\[Tau]_,ell_], \[Lambda]_, queryLst_]:=
Module[{sortedQuery,query,FixPair,sizeA,sizeS, time,
currCoupling,currPolicy,currDist,empty,
vtc,scheduleIndex, rcCell,oldSchedule, newSchedule,iterations=0, return},

FixPair[dist_,{s_,t_}]:=
	(dist[s,t] = Fixed[dist[s,t]];
	(*Unset[currCoupling[{s,t}]];*) 
	dist);

(* Initialization *)
time = First@Timing[
currDist=InitDistances[mc,OptionValue[BDistMC,Estimates]];
sortedQuery=DeleteDuplicates[Sort/@Select[queryLst,Not@FixedDistanceQ[currDist,#]&]];
currCoupling = EmptyCoupling[];
While[sortedQuery != {},
DebugPrint["Pairs left to consider: ",sortedQuery];
(* pick at random a pair from the query *)
query=sortedQuery[[{1}]];(*sortedQuery\[LeftDoubleBracket]RandomInteger[{1,Length@sortedQuery},1]\[RightDoubleBracket];*)
DebugPrint["Picked pair: ",query];
currCoupling = AddPairs[mc,currDist][currCoupling,query];
PrintActiveProblems[currDist,currCoupling,query];
currDist = UpdateDistanceValues[\[Lambda],mc,currDist,currCoupling,query];
(* look for a non-optimal transportation schedule *)
vtc = MatchingToChange[currCoupling,currDist,query];
DebugPrint["Vertex to change:",vtc];
While[vtc != {},
iterations++;
{scheduleIndex, rcCell} = vtc;(* consider the schedule to be optimized *)
(* move to a better coupling *)
oldSchedule= currCoupling[scheduleIndex];
DebugPrint["BestVertex",RelativeCostMatrix[currDist,oldSchedule], rcCell, oldSchedule];
newSchedule = BestVertex[RelativeCostMatrix[currDist,oldSchedule], rcCell, oldSchedule];
DebugPrint["NewSchedule",newSchedule];
currCoupling = UpdateCoupling[mc,currDist][currCoupling,scheduleIndex, newSchedule];
DebugPrint["Updated"];

PrintActiveProblems[currDist,currCoupling,query];
(* update the current distance with the optimal value function for the current coupling *)
currDist = UpdateDistanceValues[\[Lambda],mc,currDist,currCoupling,query];
(*DebugPrint[MatrixForm@currDist];*)
(* look for the next non-optimal transportation schedule *)
vtc = MatchingToChange[currCoupling,currDist,query];
];
(* Fix the distance for the current active pairs *)
DebugPrint["Pair done: fixing values for the following pairs",ActivePairs[currDist,currCoupling,query]];

currDist=Fold[FixPair,currDist,ActivePairs[currDist,currCoupling,query]];
(* update the query list *)
sortedQuery = Select[sortedQuery,Not@FixedDistanceQ[currDist,#]&];
]];
DebugPrint["There are no more pairs to process."];
DebugPrint["Total solved TPs: ",iterations];
DebugPrint["Execution Time: ",time, " sec"];

Switch[OptionValue[BDistMC,Output],
RulesForm, (#->getDistValue[currDist,#]&)/@queryLst,
VerboseForm, R[time, iterations, currDist],
OptimalCoupling,
(*Print[(# -> PrPrint@currCoupling@#)&/@ CouplingStates[mc,currCoupling,queryLst]];*)
CompleteCoupling[mc,currCoupling,DeleteDuplicates[Sort/@queryLst]],
_, currDist]
];


(* ::Subsection::Closed:: *)
(*Bisimilarity Quotient*)


BisimQuotientMC[mc:MC[M_,tau_,ell_]]:=
Module[{Q,btau,bell,s},

Q=BisimClassesMC[mc];
Do[
s = First[c]; (* representant of the class c *)
bell[s]=ell[s]; (* labeling function *)
(* transition function *)
btau[s]=Cases[(First[#]->Total[(#/.Append[tau[s],_->0]&)/@#])&/@Q,Except[_->0]];
(*btau[s]=Cases[(First[#]->Total[(#/.Append[tau[s],_->0]&)/@#])&/@Q,HoldPattern[_ -> v_/;v>0]];*)
,{c,Q}];

MC[First/@Q,btau,bell]
];


BisimClassesMC[mc:MC[S_,_,_]]:=
\!\(TraditionalForm\`FixedPoint\)[(Union@@#&)/@ Gather[#,Intersection[#1,#2]!={}&]&,
Select[BDistMC[mc,1/2,Tuples[S,2],Verbose->False,Output->RulesForm],#[[2]]==0&][[All,1]]];


(* ::Subsection::Closed:: *)
(*Optimal Bounded Approximant Problem (OBA)*)


ReachableCouplingStates[mc:MC[_,_,ell_],coupling_,queryLst_]:=
Module[{visitedQ}, visitedQ[_]:= False; Fold[VisitCoupling[mc,coupling],visitedQ,queryLst]];
VisitCoupling[mc:MC[_,_,ell_],coupling_][visitedQ_,pair:{s_,t_}]:=visitedQ /; visitedQ@pair \[Or] ell[s]!=ell[t];
VisitCoupling[mc:MC[_,_,_],coupling_][visitedQ_,pair:{_,_}]:=
(visitedQ[pair]=True; Fold[VisitCoupling[mc,coupling],visitedQ,NextDemandedPairs[coupling[pair]]]);


CouplingStates[mc:MC[_,_,_],coupling_,queryLst_]:= SelectVisited@ReachableCouplingStates[mc,coupling,queryLst];
(*#[[1,1,1]]&/@Select[DownValues[Evaluate@ReachableCouplingStates[mc,coupling,queryLst]],#[[2]]&];*)


(* ::Subsubsection::Closed:: *)
(*Find Locally Optimal Coupling Structure with fixed marginals*)


OptimalCouplingStruct[mcM:MC[M_,_,ell_],mcN:MC[N_,_,alpha_],\[Lambda]_,R_]:=
Module[{pairs={},bisim={},optcoupling,coupl},

(* select the pairs that have to be part of the coupling &
fix the estimate distance for the bisimilar pairs *)
Do[If[R[m,n],PrependTo[bisim,{left[m],right[n]}->0]],{m,M},{n,N}];
pairs = Join@@Table[{left[m],right[n]},{m,M},{n,N}];

optcoupling = BDistMC[JoinMC[mcM,mcN],\[Lambda],pairs,Output->OptimalCoupling,Estimates->bisim];
Do[
coupl[m,n,u,v]= If[True,(*ell[m]==alpha[n]\[And]\[Not]R[m,n]*)
getCouplingValue[optcoupling,{left[m],right[n]},{left[u],right[v]}],0]
,{m,M},{n,N},{u,M},{v,N}];
coupl
];


(* ::Subsubsection::Closed:: *)
(*Compute the E[N[n,v]]*)


ReverseReachable[mcM:MC[_,_,_],mcN:MC[_,_,_]][coupling_,queryLst_]:=
Module[{visitedQ}, visitedQ[_]:= False; Fold[ReverseVisit[mcM,mcN,coupling],visitedQ,queryLst]];
ReverseVisit[mcM:MC[_,_,_],mcN:MC[_,_,_],coupling_][visitedQ_,pair:{_,_}]:=visitedQ /;visitedQ@pair;
ReverseVisit[mcM:MC[_,_,_],mcN:MC[_,_,_],coupling_][visitedQ_,pair:{_,_}]:=
(visitedQ[pair]=True; Fold[ReverseVisit[coupling],visitedQ,ReverseAdiacentPairs[mcM,mcN,coupling,pair]]);


ReverseAdiacentPairs[mcM:MC[M_,_,_],mcN:MC[N_,_,_],coupling_,{m_,n_}]:=
Module[{allpairs,goes2mn},
goes2mn[{u_,v_}]:= coupling[u,v,m,n]>0;
Select[Join@@Table[{u,v},{u,M},{v,N}],goes2mn]
];


ComputePhi[mcM:MC[M_,_,ell_],mcN:MC[N_,_,alpha_],\[Lambda]_,coupling_,R_]:=
Module[{\[Phi],equalities,vars,sol,iota},
iota[m_,n_]:= If[MatchQ[{m,n},{First@M,First@N}],1,0];

(* \[Phi][m,n] \[Equal] 0 if \[ScriptL](m) \[NotEqual] \[Alpha](n) or (m,n) \[Element] R*)
\[Phi][m_,n_]:= x[m,n] == 0/;ell[m] != alpha[n] \[Or] R[m,n];
(* \[Phi][m,n] \[Equal] \[Iota](m,n) + \[Lambda] Underscript[\[Sum], u,v] C(u,v)(m,n)(1 + \[Phi][u,v]) otherwise *)
\[Phi][m_,n_]:= x[m,n] == iota[m,n] + \[Lambda](Join@@Table[coupling[u,v,m,n],{u,M},{v,N}].(1+vars));
(*Print[Join@@Table[f[u,v,m,n],{u,M},{v,N}].(1+vars)]*)

vars = Flatten@Table[x[m,n],{m,M},{n,N}];
equalities = Flatten@Table[\[Phi][m,n],{m,M},{n,N}];
(*Print[equalities];Abort[];*)
sol=SolveLinearEquations[equalities,vars];
(*Print["\[Phi]: ",sol];*)
Thread[Rule[Join@@Table[{m,n},{m,M},{n,N}],sol]]
];


ComputeBeta[mcM:MC[M_,_,ell_],mcN:MC[N_,_,alpha_],\[Lambda]_,coupling_,R_]:=
Module[{\[Beta],equalities,vars,sol},
iota[m_,n_]:= If[MatchQ[{m,n},initState/@{mcM,mcN}],1,0];

\[Beta][m_,n_]:= x[m,n] == 1/;R[m,n]; (* \[Beta][m,n] \[Equal] 1 if (m,n) \[Element] R*)
\[Beta][m_,n_]:= x[m,n] == 0/;ell[m] != alpha[n]; (* \[Beta][m,n] \[Equal] 0 if \[ScriptL](m) \[NotEqual] \[Alpha](n)*)
(* \[Beta][m,n] \[Equal] 1 - \[Lambda] + \[Lambda] Underscript[\[Sum], u,v] C(m,n)(u,v) \[Beta][u,v]) otherwise *)
\[Beta][m_,n_]:= x[m,n] == 1 - \[Lambda] + \[Lambda](Join@@Table[coupling[m,n,u,v],{u,M},{v,N}].vars);

vars = Join@@Table[x[m,n],{m,M},{n,N}];
equalities = Join@@Table[\[Beta][m,n],{m,M},{n,N}];
(*Print[equalities];Abort[];*)
sol = SolveLinearEquations[equalities,vars];
(*Print["\[Beta]: ",sol];*)
Thread[Rule[Join@@Table[{m,n},{m,M},{n,N}],sol]]
];


(* ::Subsection:: *)
(*EM Algorithm*)


(* ::Subsubsection:: *)
(*AveragedMarginal Heuristic*)


EN[mcM:MC[M_,_,_],mcN:MC[N_,_,_],\[Lambda]_,coupling_,R_]:=
Module[{\[Phi],\[Beta],E},
\[Phi] = ComputePhi[mcM,mcN,\[Lambda],coupling,R];
(*Print["\[Phi]: ",\[Phi]];*)
\[Beta] = ComputeBeta[mcM,mcN,\[Lambda],coupling,R];
(*Print["\[Beta]: ",\[Beta]];*)
Do[(* set expected value *)
E[m,n,u,v] = ({m,n}/.\[Phi])* \[Lambda] * coupling[m,n,u,v] * ({u,v}/.\[Beta]);
,{m,M},{n,N},{u,M},{v,N}];
E
]/;OptionValue[MinimizeMC,Method]=="AveragedMarginal";


EMAlgorithm[mcM:MC[M_,\[Tau]_,ell_],mcN:MC[N_,\[Theta]_,alpha_],\[Lambda]_,coupling_,R_]:=
Module[{newTheta,E,c,B,getVal},
getVal[theta_,x_,y_]:= y/.(Append[theta[x],_->0]);

(* calulate the expected values *)
E = EN[mcM,mcN,\[Lambda],coupling,R];
(*Print["E[N]: ", Flatten@Table[{n,v}\[Rule]E[n,v],{n,N},{v,N}]];Abort[];*)
Do[c[m,n,u,v] = If[R[m,n]||ell[m]!=alpha[n],
0,
getVal[\[Tau],m,u] * SafeDivision[E[m,n,u,v],Total[Table[E[m,n,u,x],{x,N}], Method->CompensatedSummation]]]
,{m,M},{n,N},{u,M},{v,N}];
Do[B[n,v] = Total[Table[c[m,n,u,v],{m,M},{u,M}],2,Method->CompensatedSummation],{n,N},{v,N}];

(* calculate the new marginal *)
Do[
newTheta[n]= If[Or@@Table[R[m,n],{m,M}], 
	\[Theta][n],
	Module[{tot = Total@Table[B[n,v],{v,N}]}, 
		Select[Table[v ->SafeDivision[B[n,v],tot],{v,N}],#[[2]]>0&]]
]
,{n,N}];

Clear[B,c,E];
(* the coupling after the update and the new mcN *)
MC[N,newTheta,alpha]
]/;OptionValue[MinimizeMC,Method]=="AveragedMarginal";


(* ::Subsubsection:: *)
(*Averaged Expectation Heuristic*)


EN[mcM:MC[M_,_,_],mcN:MC[N_,_,_],\[Lambda]_,coupling_,R_]:=
Module[{\[Phi],\[Beta],E},
\[Phi] = ComputePhi[mcM,mcN,\[Lambda],coupling,R];
(*Print["\[Phi]: ",\[Phi]];*)
\[Beta] = ComputeBeta[mcM,mcN,\[Lambda],coupling,R];
(*Print["\[Beta]: ",\[Beta]];*)
Do[(* set expected value *)
E[n,v] = Total[Table[({m,n}/.\[Phi])* \[Lambda] * coupling[m,n,u,v] * ({u,v}/.\[Beta]),{m,M},{u,M}],2];
,{n,N},{v,N}];
E
]/;OptionValue[MinimizeMC,Method]=="AveragedExpectation";


EMAlgorithm[mcM:MC[M_,_,_],mcN:MC[N_,\[Theta]_,alpha_],\[Lambda]_,coupling_,R_]:=
Module[{newTheta,E},

(* calulate the expected values *)
E = EN[mcM,mcN,\[Lambda],coupling,R];
(*Print["E[N]: ", Flatten@Table[{n,v}\[Rule]E[n,v],{n,N},{v,N}]];Abort[];*)
(* calculate the new marginal *)
Do[
newTheta[n]= If[Or@@Table[R[m,n],{m,M}], 
	\[Theta][n],
	Module[{tot = Total[Table[E[n,v],{v,N}],Method->CompensatedSummation]}, 
		Select[Table[v ->SafeDivision[E[n,v],tot],{v,N}],#[[2]]>0&]]
]
,{n,N}];

(* the coupling after the update and the new mcN *)
MC[N,newTheta,alpha]
]/;OptionValue[MinimizeMC,Method]=="AveragedExpectation";


(* ::Subsubsection:: *)
(*Minimize Markov Chain*)


Bisimilarity[mcM:MC[M_,_,_],mcN:MC[N_,_,_]]:=
Module[{R,f,pairs},
R[_,_]:=False;
f[{{_,m_},{_,n_}}-> 0]:= (R[m,n]=True);
pairs = Join@@Table[{left[m],right[n]},{m,M},{n,N}];
Scan[f,BDistMC[JoinMC[mcM,mcN],0.5,pairs,Verbose->False,Output->RulesForm]];
R
];


Options[MinimizeMC]={Output->Automatic,Verbose->False,MaxIter->20,Timeout->600,Method->"AveragedExpectation"};
MinimizeMC::Timeout ="The execution has exceeded the given timeout of "<>ToString@OptionValue[MinimizeMC,Timeout]<>" seconds";
MinimizeMC::MaxIter= "Interation Limit Exceeded. Use the 'MaxIter' option to set a custom iteration limit.";


MinimizeMC[mcM:MC[_,_,_],mcN:MC[_,_,_],\[Lambda]_,
options:OptionsPattern[{Output->Automatic,Verbose->False,MaxIter->20,Timeout->600,Method->"AveragedExpectation"}]]:=
TimeConstrained[
SetOptions[MinimizeMC,Verbose->OptionValue[Verbose]];
SetOptions[MinimizeMC,Method->OptionValue[Method]];
Module[{R = Bisimilarity[mcM,mcN],
dist,currmc,time,olddist,currdist,test,f,iter=0,list,logg={}},

dist[mc_]:= BDistMC[JoinMC[mcM,mc],\[Lambda],{{left@initState@mcM,right@initState@mc}}][[1,2]];

f[{mc_,_}]:= Module[{chain},
iter++;
chain = EMAlgorithm[mcM,mc,\[Lambda],OptimalCouplingStruct[mcM,mc,\[Lambda],R],R];
{chain,dist[chain]}
];

test[{_,d1_},{_,d2_}]:= d1 > d2;

olddist = dist[mcN]; (* Distance from the inital MC *)
{time,{currmc,currdist}} = Timing[
list = NestWhileList[f,{mcN,olddist},test,2,OptionValue[MinimizeMC,MaxIter]];
If[iter >= OptionValue[MinimizeMC,MaxIter]&& test@@Take[list,-2],
Message[MinimizeMC::MaxIter]; list[[-1]], 
list[[-2]]]
];
(* precmc and precdist are the best encountered so far *)
(*OptionValue[MinimizeMC,MaxIter]*)
AppendTo[logg,"Time"->time]; 
AppendTo[logg,"Iterations"->iter];
AppendTo[logg,"dList"->(#[[2]]&/@list)];
AppendTo[logg,"d(M,N0)"-> olddist];
AppendTo[logg,"d(M,NL)"-> currdist];

If[OptionValue[MinimizeMC,Verbose], 
Print["Time Used: ",time, ", Iterations: ", iter];
Print["Steps: \n",TableForm[{PlotMC[#[[1]]],#[[2]]}&/@list]];
Print["Distance from the inital MC: ",olddist];
Print["Distance from the final MC: ",currdist]
];
If[MatchQ[OptionValue[MinimizeMC,Output],"Stats"],
logg,currmc](*ChopProbabilities@*)
],OptionValue[MinimizeMC,Timeout],Message[MinimizeMC::Timeout];Abort[]];


End[]


EndPackage[]
